/************************\
*                        *
*  Implements TCADSheet  *
*                        *
\************************/

#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include "sheet.h"
#include "texteditbox.h"
#include "messagebox.h"
#include "main.h"
#include "rightangleline.h"
#include "simpleline.h"
#include "text.h"
#include "frame.h"
#include "framedtext.h"
#include "gridoptions.h"
#include "canvassize.h"
#include "epsheader.h"
#include "messagedlg.h"

/*
 * Initial size of canvas
 */
#define INITIAL_WIDTH 1000
#define INITIAL_HEIGHT 1000

// Buffer
static char Buf[PATH_MAX+1024];

// User data for EditText
struct TEditTextData {
	TCADSheet *Sheet;
	gboolean ModifyOnCancel;
};

// Alignment macro
#define ALIGN(x,gran,Force) \
if ((UseGrid) || (Force)) { \
  int _T = x % gran; \
  if (_T < (gran/2)) x -= _T; /* Round down */ \
  else x += (gran-_T); /* Round up  */\
}

/* Callback to set sheet size from dialog */
static void SetSheetSizeCB(TCADSheetSize &size,TCADSheet *sheet)
{
	DTP("SetSheetSizeCB()\n");

	sheet->SetSheetSize(size);
}

/* Callback for "save on close" question */
static void SaveOrNotCB(int Button,TCADSheet *S)
{
	DTP("SaveOrNotCB()\n");

	if (Button == 0) System->Save(S); // Yes
	else if (Button == 1) {
		// No
		S->Saved();	// !!! Trick !!!
		System->CloseSheet(S);
		if (System->ShuttingDown) System->Menu_Exit();
	}
	else { // Cancel
		S->Closing=false; 
		System->ShuttingDown=false;
	}
}

/*
 * Class implementation
 */

#define STD_MASK GDK_EXPOSURE_MASK+GDK_BUTTON_PRESS_MASK+GDK_BUTTON_RELEASE_MASK+GDK_VISIBILITY_NOTIFY_MASK

TCADSheet::TCADSheet(char Visible) : TCADGroup(this),size()
{
	DTP("TCADSheet::TCADSheet()\n");

	BlockChange=1;

  Uniq = System->NextUniq();

  Sheet = this;
  FirstSelected = NULL;
  FileName = NULL;
  FirstView = NULL;
  LastUniq = 0;
	Closing=false;
	Toggling=false;

  // Init grid
  UseGrid = TRUE;
  ShowGrid = TRUE;
  GX_eq_GY = TRUE;
  GridX = 8;
  GridY = 8;

  // Set EncapRect
  EncapRect.x = 0;
  EncapRect.y = 0;
  EncapRect.width = INITIAL_WIDTH;
  EncapRect.height = INITIAL_HEIGHT;

	SavedCount=0;
	ChangeCount=0;

	bzero(UndoMem,MAX_UNDO_LEVELS*sizeof(TCADMemory *));

  // Create first view
  new TCADView(this,Visible);

	UMAct = -1;
	UMFirst = 0;
	UMLast = 0;
	if (Visible) SaveState();

	BlockChange=0;
}

TCADSheet::~TCADSheet()
{
  TCADView *View;

	DTP("TCADSheet::~TCADSheet()\n");

	Closing=1;
	
  // Destroy window
//  printf("Closing!\n");

  // Destroy all views
  while(FirstView) {
    View = FirstView;
    FirstView = View->Next;
    View->Sheet = NULL;
    delete View;
  }

  if (FileName != NULL) free(FileName);
}

void
TCADSheet::Save(TCADSaveStream *Fl)
{
	DTP("TCADSheet::Save()\n");

  Fl->SaveObjectName("TCADSheet");
  SaveProperties(Fl);
  Fl->SaveEndObject();
}

void
TCADSheet::SaveProperties(TCADSaveStream *Fl)
{
	int X,Y,W,H;

	DTP("TCADSheet::SaveProperties()\n");

  GtkWidget *Window = GTK_WIDGET(FirstView->GetGtkWindow());

  TCADGroup::SaveProperties(Fl);

	gdk_window_get_position(Window->window,&X,&Y);
	gdk_window_get_size(Window->window,&W,&H);
	WS.x = X;
	WS.y = Y;
	WS.width = W;
	WS.height = H;

	if (Fl->SaveOnlySelected) return;

  Fl->SaveInt("WndX",WS.x);
  Fl->SaveInt("WndY",WS.y);
  Fl->SaveInt("WndW",WS.width);
  Fl->SaveInt("WndH",WS.height);

  Fl->SaveInt("LastUniq",LastUniq);

  Fl->SaveInt("UseGrid",UseGrid);
  Fl->SaveInt("ShowGrid",ShowGrid);
  Fl->SaveInt("GX_eq_GY",GX_eq_GY);
  Fl->SaveInt("GridX",GridX);
  Fl->SaveInt("GridY",GridY);

  Fl->SaveInt("HRulerVisible",FirstView->HRulerSettings.Visible);
  Fl->SaveInt("VRulerVisible",FirstView->VRulerSettings.Visible);
  Fl->SaveInt("HRulerMetric",FirstView->HRulerSettings.Metric);
  Fl->SaveInt("VRulerMetric",FirstView->VRulerSettings.Metric);
	size.Save(Fl);
}

void
TCADSheet::PostLoad(TCADLoadStream *Fl)
{
	DTP("TCADSheet::PostLoad()\n");

  TCADGroup::PostLoad(Fl);

	ApplySheetSize();
  FirstView->ApplyRulerSettings();
	FirstView->SetupCanvas(EncapRect.width, EncapRect.height);
	FirstView->SetupWindow(WS.x, WS.y, WS.width, WS.height);

	if (!Fl->KeepChanges) {
		SavedCount=0;
		ChangeCount=0;
		BlockChange=0;

		UMFirst = 0;
		UMLast = 0;
		UMAct = -1;
	}
	else BlockChange--;	// Was incremented in Load()
}

void
TCADSheet::Load(TCADLoadStream *Fl)
{
	DTP("TCADSheet::Load()\n");

	BlockChange++;

	TCADGroup::Load(Fl);
}

char
TCADSheet::LoadProperty(TCADLoadStream *Fl,char *ID)
{
  char Result=1;

	DTP("TCADSheet::LoadProperty()\n");

  if (strcasecmp("WndX",ID) == 0) WS.x = Fl->LoadInt();
  else if (strcasecmp("WndY",ID) == 0) WS.y = Fl->LoadInt();
  else if (strcasecmp("WndW",ID) == 0) WS.width = Fl->LoadInt();
  else if (strcasecmp("WndH",ID) == 0) WS.height = Fl->LoadInt();
  else if (strcasecmp("LastUniq",ID) == 0) LastUniq = Fl->LoadInt();
  else if (strcasecmp("UseGrid",ID) == 0) UseGrid = Fl->LoadInt();
  else if (strcasecmp("ShowGrid",ID) == 0) ShowGrid = Fl->LoadInt();
  else if (strcasecmp("GX_eq_GY",ID) == 0) GX_eq_GY = Fl->LoadInt();
  else if (strcasecmp("GridX",ID) == 0) GridX = Fl->LoadInt();
  else if (strcasecmp("GridY",ID) == 0) GridY = Fl->LoadInt();
  else if (strcasecmp("HRulerVisible",ID) == 0) FirstView->HRulerSettings.Visible = Fl->LoadInt();
  else if (strcasecmp("VRulerVisible",ID) == 0) FirstView->VRulerSettings.Visible = Fl->LoadInt();
  else if (strcasecmp("HRulerMetric",ID) == 0) FirstView->HRulerSettings.Metric = (GtkMetricType)Fl->LoadInt();
  else if (strcasecmp("VRulerMetric",ID) == 0) FirstView->VRulerSettings.Metric = (GtkMetricType)Fl->LoadInt();
	else if (strcasecmp("TCADSheetSize",ID) == 0) size.Load(Fl);
  else Result = TCADGroup::LoadProperty(Fl,ID);
  
  return(Result);
}

gboolean
TCADSheet::Close()
{
	DTP("TCADSheet::Close()\n");

	Closing=true;

	if (!IsModified()) return(true);

	// Ask whether we should save
	sprintf(Buf,"File '%s' has been modified since last save. Do you want to save it?",GetTitle());
	MessageDlg(Buf,"Question","Yes|No|Cancel",SaveOrNotCB,this);
	return(false);
}

void
TCADSheet::DrawRect(TCADView *View,GdkRectangle *ORect,int MagA,int MagB)
{
  GdkRegion *Region;
  GdkPoint Points[5];
  TCADObject *Child;
  TCADHandle *Handle;
  int i,x,y,x2,y2;
  GdkDrawable *d;
	int s;
	GdkRectangle Rect;
	
//	DTP("TCADSheet::DrawRect(,,,)\n");

	d = View->GetDrawable();
	if (d == NULL) return;

/*
  if (Rect != NULL) printf("TCADSheet::DrawRect(%d,%d,%d,%d)\n",Rect->x,Rect->y,Rect->width,Rect->height);
  else printf("DrawRect:NULL\n");
*/

  /* Create region */
  if (ORect == NULL) {
    Region = NULL;
    Rect = EncapRect;

    // Clear area
    View->Clear(&Rect,MagA,MagB);
  }
  else {
    // Limit rectangle
    x = ORect->x;
    y = ORect->y;
    x2 = ORect->x+ORect->width;
    y2 = ORect->y+ORect->height;

		// Enlarge rectangle by current handle size...
		s = HANDLE_SIZE/2*MagB/MagA;
		x -= s;
		y -= s;
		x2 += s;
		y2 += s;
	
    if (x < 0) x = 0;
    if (y < 0) y = 0;
    if (x2 < 0) x2 = 0;
    if (y2 < 0) y2 = 0;

    if (x > EncapRect.width) x = EncapRect.width;
    if (y > EncapRect.height) y = EncapRect.height;
    if (x2 > EncapRect.width) x2 = EncapRect.width;
    if (y2 > EncapRect.height) y2 = EncapRect.height;
    
    Rect.x = x;
    Rect.y = y;
    Rect.width = x2-x;
    Rect.height = y2-y;
    
    // Fill points
    Points[0].x = Rect.x;
    Points[0].y = Rect.y;
    Points[1].x = Rect.x+Rect.width;
    Points[1].y = Rect.y;
    Points[2].x = Rect.x+Rect.width;
    Points[2].y = Rect.y+Rect.height;
    Points[3].x = Rect.x;
    Points[3].y = Rect.y+Rect.height;
    Points[4].x = Rect.x;
    Points[4].y = Rect.y;

    Region = gdk_region_polygon(Points,5,GDK_EVEN_ODD_RULE);

    // Set clipping, draw grid and clear
    View->Clear(&Rect,MagA,MagB);
  }

  Draw(View,Region,MagA,MagB);    // TCADGroup::Draw
  View->Draw(MagA,MagB);		// Let view draw what it wants (i.e., focus rectangle)

  // Draw handles (if any)
	gdk_gc_set_line_attributes(View->GC,1,GDK_LINE_SOLID,GDK_CAP_NOT_LAST,GDK_JOIN_MITER);
  Child = FirstSelected;
  while(Child) {
    Handle = Child->Handles;
    if (Handle) {
      for(i=Child->NumHandles;i>0;i--) {
        Handle->Draw(View,View->GC,(Child==FirstSelected),MagA,MagB);
        Handle++;
      }
    }
     
    Child = Child->NextSelected;
  }

  if (Region != NULL) gdk_region_destroy(Region);
}

unsigned long
TCADSheet::NextUniq()
{
	DTP("TCADSheet::NextUniq()\n");

  return(++LastUniq);
}

void
TCADSheet::SelectObject(TCADObject *Object,GdkRectangle *RefreshRect,GdkRectangle *RefreshRect2)
{
  TCADObject *PrevFirst;

	DTP("TCADSheet::SelectObject()\n");

  RefreshRect->x = 0;
  RefreshRect->y = 0;
  RefreshRect->width = 0;
  RefreshRect->height = 0;
  RefreshRect2->x = 0;
  RefreshRect2->y = 0;
  RefreshRect2->width = 0;
  RefreshRect2->height = 0;
  
  if (Object->Selected) {
    return;
  }
  // Enqueue currently selected object
  PrevFirst = FirstSelected;
  Object->PrevSelected = NULL;
  Object->NextSelected = FirstSelected;
  if (FirstSelected) FirstSelected->PrevSelected = Object;

  FirstSelected = Object;
  Object->Select(RefreshRect);

  // Do OtherSelected()
  if (PrevFirst != NULL) PrevFirst->OtherSelected(RefreshRect2);
}

TCADLine *
TCADSheet::CreateLine(int Type,int x,int y)
{
  TCADLine *Line;

	DTP("TCADSheet::CreateLine()\n");

  if (System->GetState() == ST_LINE) Line = new TCADSimpleLine(this,x,y);
  else Line = new TCADRightAngleLine(this,x,y);
  Line->Uniq = NextUniq();
 
  AppendChild(Line); 

  // Make line selected
  Line->PrevSelected = NULL;
  Line->NextSelected = NULL;
  FirstSelected = Line;
  Line->Selected = TRUE;

	Changed();
  return(Line);
}


TBPResult
TCADSheet::ButtonPressed(int x,int y,gboolean Shift,TCADView *View)
{
  TCADObject *Selected,*Child;
  GdkRectangle RefreshRect,RefreshRect2;
  char WasFirst;
  char UnsAll;
  TCADHandle *Handle;
  int i;
  TCADLine *Line;
  int MinX,MinY;
	int d;
	int MagA,MagB;

	DTP("TCADSheet::ButtonPressed()\n");

  // Check handles
  Handle = NULL;
  i = 0;
  if (FirstSelected) {
		View->GetMag(MagA,MagB);
		d = (HANDLE_SIZE/2 + 1)*MagB/MagA;

    Handle = FirstSelected->Handles;
    for(i=0;i<FirstSelected->NumHandles;i++) {
      if ((x >= (Handle->x-d)) && (x < (Handle->x+d)) &&
          (y >= (Handle->y-d)) && (y < (Handle->y+d))) break;
      Handle++;
    }
    if (i == FirstSelected->NumHandles) Handle = NULL;
  }

  if (Handle != NULL) {
    // Handle clicked
    FirstSelected->HandleButtonPressed(i,x,y);
    View->HandleClicked(x,y,Handle);
    return(BPR_HANDLE);
  }

  Selected = TCADGroup::ButtonPressed(x,y);

//  if ((Selected == NULL) && (System->GetState() == ST_PICK)) return(FALSE);

//  printf("Selected=%x, Shift=%d\n",(int)Selected,Shift);
  UnsAll = !Shift;
  if ((UnsAll) && (Selected != NULL))
   if (Selected->Selected) UnsAll = 0;

  if (UnsAll) UnselectAll();

  if (Selected) {
    if ((Selected->Selected) && (Shift)) {
      // Remove object from list
      WasFirst = (Selected == FirstSelected);
      if (Selected->PrevSelected == NULL) FirstSelected = Selected->NextSelected;
      else Selected->PrevSelected->NextSelected = Selected->NextSelected;
      if (Selected->NextSelected != NULL)
       Selected->NextSelected->PrevSelected = Selected->PrevSelected;

      // Unselect current object
      Selected->Unselect(&RefreshRect);

      // Redraw
      DrawRect(&RefreshRect);

      // Re-select first item
      if ((FirstSelected) && (WasFirst)) {
        FirstSelected->Select(&RefreshRect);

        // Redraw
        DrawRect(&RefreshRect);
      }
    }
    else if (Selected != FirstSelected) {
      SelectObject(Selected,&RefreshRect,&RefreshRect2);

      // Redraw
      DrawRect(&RefreshRect);
      if (RefreshRect2.width > 0) DrawRect(&RefreshRect2);
    }
  }
  
  if (FirstSelected) {
    // Compute minimal X
    MinX = FirstSelected->EncapRect.x;
    MinY = FirstSelected->EncapRect.y;
    Child = FirstSelected->NextSelected;
    while(Child) {
      MinX = MIN(Child->EncapRect.x,MinX);
      MinY = MIN(Child->EncapRect.y,MinY);
      Child=Child->NextSelected;
    }
  
    // Compute corrections
    View->XCorr = MinX-x;
    View->YCorr = MinY-y;
  }

	if (Selected) return(BPR_OBJECT);
	else return(BPR_NOTHING);
}

void
TCADSheet::SelectRectangle(int x1,int y1,int x2,int y2)
{
  GdkRectangle Rect1,Rect2,Rect3,Rect4,*ER;
  gboolean First;
  TCADObject *Obj;

	DTP("TCADSheet::SelectRectangle()\n");

  First = TRUE;
  Rect1.x = x1;
  Rect1.y = y1;
  Rect1.width = x2-x1+1;
  Rect1.height = y2-y1+1;

  Obj = FirstChild;
  while(Obj) {
    if (!Obj->Selected) {
      ER = &Obj->EncapRect;
      if ((ER->x >= x1) && ((ER->x+ER->width ) <= x2) &&
          (ER->y >= y1) && ((ER->y+ER->height) <= y2)) {
        if (First) SelectObject(Obj,&Rect3,&Rect2);
        else SelectObject(Obj,&Rect3,&Rect4);
        System->UnionRectangles(&Rect1,&Rect3);
        First = TRUE;
      }
    }
    Obj = Obj->Next;
  }

  // Redraw
  DrawRect(&Rect1);
  if (!First) {
    if (Rect2.width > 0) DrawRect(&Rect2);
  }
}

void
TCADSheet::NewRectObject(int Kind,GdkRectangle *Bounds)
{
  GdkRectangle Rect1,Rect2;
  TCADObject *Obj;

	DTP("TCADSheet::NewRectObject()\n");

  switch(Kind) {
    case ST_FRAME:
      Obj = new TCADFrame(this);
      Obj->Uniq = NextUniq();
      AppendChild(Obj);
      Obj->EncapRect = *Bounds;
      Obj->MoveResize(0,0,0,0,&Rect1);
      DrawRect(&Rect1);
      SelectObject(Obj,&Rect1,&Rect2);
      DrawRect(&Rect1);
			Changed();
      break;
    case ST_TEXT:
      Obj = new TCADText(this);
      Obj->Uniq = NextUniq();
      AppendChild(Obj);
      Obj->EncapRect = *Bounds;
      Obj->MoveResize(0,0,0,0,&Rect1);
      DrawRect(&Rect1);
      SelectObject(Obj,&Rect1,&Rect2);
      DrawRect(&Rect1);
      EditText(true);
      break;
    case ST_FRAMED_TEXT:
      Obj = new TCADFramedText(this);
      Obj->Uniq = NextUniq();
      AppendChild(Obj);
      Obj->EncapRect = *Bounds;
      Obj->MoveResize(0,0,0,0,&Rect1);
      DrawRect(&Rect1);
      SelectObject(Obj,&Rect1,&Rect2);
      DrawRect(&Rect1);
      EditText(true);
      break;
    default: ;
  }

  return;
}

void
TCADSheet::MoveSelected(int dx,int dy)
{
  GdkRectangle Rect1,Rect2;
  char First;
  TCADObject *Obj;

	DTP("TCADSheet::MoveSelected()\n");

	if ((dx == 0) && (dy == 0)) return;

	BlockChange++;

  // Move dragged objects
  Obj = FirstSelected;
  First = 1;
  while(Obj) {
    if (First) FirstSelected->MoveResize(dx,dy,0,0,&Rect1);
    else {
      Obj->MoveResize(dx,dy,0,0,&Rect2);
      System->UnionRectangles(&Rect1,&Rect2);
    }

    First = 0;
    Obj = Obj->NextSelected;
  }
  
  if (!First) {
    // Redraw
    DrawRect(&Rect1);
  }

	BlockChange--;
	Changed();
}

void
TCADSheet::UnselectAll()
{
  TCADObject *Child;
  char First;
  GdkRectangle RefreshRect,RefreshRect2;

	DTP("TCADSheet::UnselectAll()\n");

  // Unselect all objects
  First = 1;
  Child = FirstSelected;
  while(Child) {
    if (First) Child->Unselect(&RefreshRect);
    else {
      Child->Unselect(&RefreshRect2);
	
      // Union rectangles
      System->UnionRectangles(&RefreshRect,&RefreshRect2);
    }
    First = 0;
    Child=Child->NextSelected;
  }
  FirstSelected = NULL;

  if (!First) DrawRect(&RefreshRect);
}

void
TCADSheet::DeleteSelected()
{
  GdkRectangle Rect;
  TCADObject *Obj,*Next;
  gboolean First = TRUE;

	DTP("TCADSheet::DeleteSelected()\n");

  Obj = FirstSelected;
  while(Obj) {
    if (First) Rect = Obj->EncapRect;
    else System->UnionRectangles(&Rect,&Obj->EncapRect);
    First = FALSE;
    
    Next = Obj->NextSelected;
    DeleteChild(Obj);
    delete Obj;
    Obj = Next;
  }
  FirstSelected = NULL;

  if (!First) {
    // Enlarge Rect to contain handles and arrows
    Rect.x -= HANDLE_SIZE/2+ARROW_SIZE_BIG/2;
    Rect.y -= HANDLE_SIZE/2+ARROW_SIZE_BIG/2;;
    Rect.width += HANDLE_SIZE+ARROW_SIZE_BIG+1;
    Rect.height += HANDLE_SIZE+ARROW_SIZE_BIG+1;
    DrawRect(&Rect);
  }

	Changed();
}

static void
EditTextCallback(char *NewText,TEditTextData *Data)
{
  TCADObject *Obj;

	DTP("EditTextCallback()\n");

  Obj = Data->Sheet->GetFirstSelected();
  if (Obj == NULL) {
		delete Data;
		return;
	}
  if (!Obj->HasText()) {
		delete Data;
		return;
	}

  if (NewText != NULL) {
    ((TCADTextObject *)Obj)->SetText(NewText);

    // Redraw object
    Data->Sheet->DrawRect(&Obj->EncapRect);

		Data->Sheet->Changed();
  }
	else {
		if (Data->ModifyOnCancel) Data->Sheet->Changed();
	}

	delete Data;
}

void
TCADSheet::EditText(gboolean ModifyOnCancel)
{
  TCADObject *Obj;
  char *OldText;
	TEditTextData *Data;

	DTP("TCADSheet::EditText()\n");

  Obj = FirstSelected;
  if (Obj == NULL) return;
  if (!Obj->HasText()) return;

	Data = new TEditTextData();
	Data->Sheet = this;
	Data->ModifyOnCancel = ModifyOnCancel;
  OldText = ((TCADTextObject *)Obj)->GetText();
  TextEditBox(OldText,NULL,EditTextCallback,Data);
  free(OldText);
}

TCADObject *
TCADSheet::GetFirstSelected()
{
//	DTP("TCADSheet::GetFirstSelected()\n");

  return(FirstSelected);
}

void
TCADSheet::Export(char *FileName,TExportFormat Format)
{
  FILE *Fl;

	DTP("TCADSheet::Export()\n");

  Fl = fopen(FileName,"w");
  if (Fl == 0) {
    char Msg[1024];
    sprintf(Msg,"Cannot open file: %s",strerror(errno));

    MessageBox(Msg,"Error");
  }
  else {
    // Write header
    switch(Format) {
		case EF_LATEX:
			fprintf(Fl,"{%% Picture saved by %s %s\n",PACKAGE,VERSION);
			fprintf(Fl,"\\unitlength=1.000000pt\n");
			fprintf(Fl,"\\begin{picture}(%d,%d)(0.00,0.00)\n",EncapRect.width,EncapRect.height);
			break;
		case EF_FIG:
			fprintf(Fl,"#FIG 3.2\n");
			fprintf(Fl,"Portrait\n");	// Orientation
			fprintf(Fl,"Center\n");		// Justification
			fprintf(Fl,"Metric\n");		// Units
			fprintf(Fl,"A4\n");		// Paper size
			fprintf(Fl,"72.00\n");		// Magnification
			fprintf(Fl,"Single\n");		// Multiple page
			fprintf(Fl,"0\n");		// Transparent color
			fprintf(Fl,"85 2\n");		// Resolution / origin UL
//	fprintf(Fl,"\n");		// 
			break;
		case EF_EPS:
			fprintf(Fl,EPS_HEADER,0,0,EncapRect.width,EncapRect.height);
		}

    TCADGroup::Export(Fl,Format);

    if (Format == EF_LATEX) {
      // Write footer
      fprintf(Fl,"\\end{picture}}\n");
    }
    fclose(Fl);
  }
}

void
TCADSheet::ShowArrowProperties()
{
  TCADObject *Obj;

	DTP("TCADSheet::ShowArrowProperties()\n");

  Obj = FirstSelected;
  while(Obj) {
    if (Obj->IsLine()) {
      ((TCADLine *)Obj)->ShowArrowProperties();
      return;
    }
    Obj = Obj->NextSelected;
  }
}

void
TCADSheet::ShowTextProperties()
{
  TCADObject *Obj;

	DTP("TCADSheet::ShowTextProperties()\n");

  Obj = FirstSelected;
  while(Obj) {
    if (Obj->HasText()) {
      ((TCADTextObject *)Obj)->ShowTextProperties(Obj);
      return;
    }
    Obj = Obj->NextSelected;
  }
}

void
TCADSheet::ShowLineProperties()
{
  TCADObject *Obj;

	DTP("TCADSheet::ShowLineProperties()\n");

  Obj = FirstSelected;
  while(Obj) {
    if (Obj->HasLineProperties()) {
			Obj->ShowLineProperties(this);
      return;
    }
    Obj = Obj->NextSelected;
  }
}
void
TCADSheet::SetShowGrid(gboolean Value)
{
	TCADView *View;

	DTP("TCADSheet::SetShowGrid()\n");

  ShowGrid = Value;
	View = FirstView;
	while(View) {
		View->RefreshCheckItems();
		View = View->Next;
	}
//  printf("TCADSheet::SetShowGrid(%d)\n",Value);
  DrawRect(NULL);
	Sheet->Changed();
}

void
TCADSheet::SetUseGrid(gboolean Value)
{
	TCADView *View;

	DTP("TCADSheet::SetUseGrid()\n");

  UseGrid = Value;
	View = FirstView;
	while(View) {
		View->RefreshCheckItems();
		View = View->Next;
	}
//  System->SheetActivated(this);
	Sheet->Changed();
}

void
TCADSheet::ShowGridProperties()
{
	DTP("TCADSheet::ShowGridProperties()\n");

  ShowGridOptionsDialog(this);
}

void
TCADSheet::GetGridOptions(gboolean& X_eq_Y,int& GX,int& GY)
{
//	DTP("TCADSheet::GetGridOptions()\n");

  X_eq_Y = GX_eq_GY;
  GX = GridX;
  GY = GridY;
}

void
TCADSheet::SetGridOptions(gboolean X_eq_Y,int GX,int GY)
{
	DTP("TCADSheet::SetGridOptions()\n");

  GX_eq_GY = X_eq_Y;
  GridX = GX;
  GridY = GY;
  DrawRect(NULL);

	Changed();
}

void
TCADSheet::AlignObjects(THorizAlign HA, TVertAlign VA)
{
  int Val1,Val2;
  TCADObject *Obj;
  int DX,DY,DW,DH,T;
  GdkRectangle Rect;

	DTP("TCADSheet::AlignObjects()\n");

  if (FirstSelected == NULL) return;

  switch(HA) {
    case HAL_LEFTS:
      Val1 = FirstSelected->EncapRect.x;
      break;
    case HAL_RIGHTS:
      Val1 = FirstSelected->EncapRect.x+FirstSelected->EncapRect.width;
      break;
    case HAL_SAMEWIDTH:
      Val1 = FirstSelected->EncapRect.width;
      break;
    default: ;
  }
  
  switch(VA) {
    case VAL_TOPS:
      Val2 = FirstSelected->EncapRect.y;
      break;
    case VAL_BOTTOMS:
      Val2 = FirstSelected->EncapRect.y+FirstSelected->EncapRect.height;
      break;
    case VAL_SAMEHEIGHT:
      Val2 = FirstSelected->EncapRect.height;
      break;
    default: ;
  }

  Obj = FirstSelected;
  while(Obj) {
    DY = 0;
    DX = 0;
    DW = 0;
    DH = 0;
    if (!Obj->IsLine()) {
			switch(HA) {
			case HAL_GRID:
				T = Obj->EncapRect.x;
				ALIGN(T,GridX,TRUE);
				DX += T-Obj->EncapRect.x;
				break;
			case HAL_LEFTS:
				DX = Val1 - Obj->EncapRect.x;
				break;
			case HAL_RIGHTS:
				DX = Val1 - (Obj->EncapRect.x+Obj->EncapRect.width);
				break;
			case HAL_SAMEWIDTH:
				DW = Val1 - Obj->EncapRect.width;
				break;	  
			default: ;
			}
      
 			switch(VA) {
			case VAL_GRID:
				T = Obj->EncapRect.y;
				ALIGN(T,GridY,TRUE);
				DY += T-Obj->EncapRect.y;
				break;
			case VAL_TOPS:
				DY = Val2 - Obj->EncapRect.y;
				break;
			case VAL_BOTTOMS:
				DY = Val2 - (Obj->EncapRect.y+Obj->EncapRect.height);
				break;
			case VAL_SAMEHEIGHT:
				DH = Val2 - Obj->EncapRect.height;
				break;	  
			default: ;
      }

      if ((DX != 0) || (DY != 0) || (DW != 0) || (DH != 0)) Obj->MoveResize(DX,DY,DW,DH,&Rect);
    }
    
    Obj = Obj->NextSelected;
  }

  // Refresh
  DrawRect(NULL);
	Changed();
}

void
TCADSheet::SelectAll()
{
  GdkRectangle Rect1;
  TCADObject *Obj;

	DTP("TCADSheet::SelectAll()\n");

  Obj = FirstChild;
  while(Obj) {
    if (!Obj->Selected) SelectObject(Obj,&Rect1,&Rect1);
    Obj = Obj->Next;
  }

  // Refresh all
  DrawRect(NULL);
}

char *
TCADSheet::GetTitle()
{
	static char TitleBuf[100];

	DTP("TCADSheet::GetTitle()\n");

  if (FileName == NULL) {
		sprintf(TitleBuf,"Untitled %d",Uniq);
		return(TitleBuf);
	}
	else return(FileName);
}

void
TCADSheet::SetTitle()
{
  char *Title;
 	TCADView *View;
 
	DTP("TCADSheet::SetTitle()\n");

	if (!FirstView) return;
	
	strcpy(Buf,GetTitle());
	
  Title = Buf;
  if (FirstView->Next == NULL) {
		if (IsModified()) {
			char Buf2[30];
			sprintf(Buf2," (modified %d)",ChangeCount-SavedCount);
			strcat(Title,Buf2);
		}
		FirstView->SetTitle(Title);
	}
  else {
    char *TmpPtr;
    int i;

    TmpPtr = Title;
    while(*TmpPtr) TmpPtr++;

    i = 1;
    View = FirstView;
    while(View) {
      sprintf(TmpPtr,":%d",i);
			if (IsModified()) strcat(TmpPtr," (modified)");

      View->SetTitle(Title);
      i++;
      View = View->Next;
    }
  }
}

void
TCADSheet::AddView(TCADView *View)
{
	DTP("TCADSheet::AddView()\n");

  if (FirstView == NULL) {
    FirstView = View;
    View->Prev = NULL;
    View->Next = NULL;
  }
  else {
    TCADView *View2;
    View2 = FirstView;
    while(View2->Next) View2 = View2->Next;
    View2->Next = View;
    View->Prev = View2;
    View->Next = NULL;
  }
  SetTitle();	// Refresh titles
}

gboolean
TCADSheet::RemoveView(TCADView *View)
{
	DTP("TCADSheet::RemoveView()\n");

	if ((View == FirstView) && (View->Next == NULL)) return(Close()); // We will close while sheet

	// Remove view
  if (View->Prev == NULL) FirstView = View->Next;
  else View->Prev->Next = View->Next;
  if (View->Next != NULL) View->Next->Prev = View->Prev;

	delete View;

	SetTitle();	// Refresh titles

  return(false);
}

gboolean
TCADSheet::HasFocus()
{
  TCADView *View;

	DTP("TCADSheet::HasFocus()\n");

  View = FirstView;
  while(View) {
    if (View->HasFocus()) return(TRUE);
    View = View->Next;
  }
  return(FALSE);
}

void
TCADSheet::DrawRect(GdkRectangle *Rect)
{
  TCADView *View;
  int MagA,MagB;

//	DTP("TCADSheet::DrawRect()\n");

//  printf("DrawRect(%d,%d,%d,%d)\n",Rect->x,Rect->y,Rect->width,Rect->height);

  View = FirstView;
  while(View) {
    View->GetMag(MagA,MagB);
    DrawRect(View,Rect,MagA,MagB);
    View = View->Next;
  }
}

void
TCADSheet::Show()
{
  TCADView *View;

	DTP("TCADSheet::Show()\n");

	BlockChange++;

  View = FirstView;
  while(View) {
    View->Show();
    View = View->Next;
  }

	if (UMAct == -1) SaveState();	// After load

	BlockChange--;
}

void
TCADSheet::Hide()
{
  TCADView *View;

	DTP("TCADSheet::Hide()\n");

  View = FirstView;
  while(View) {
    View->Hide();
    View = View->Next;
  }
}

void
TCADSheet::SetSheetSize(TCADSheetSize &size)
{
	DTP("TCADSheet::SetSheetSize()\n");

	this->size = size;
	ApplySheetSize();
}

void
TCADSheet::ApplySheetSize()
{
	TCADView *v;

	DTP("TCADSheet::ApplySheetSize()\n");

	EncapRect.width = size.width;
	EncapRect.height = size.height;

	v = FirstView;
	while(v) {
		v->SetupCanvas(size.width,size.height);
		v = v->Next;
	}

	// Refresh
	DrawRect(NULL);
	Changed();
}

void
TCADSheet::GetSheetSize(int &Width,int &Height)
{
	DTP("TCADSheet::GetSheetSize()\n");

	Width = size.width;
	Height = size.height;
}

void
TCADSheet::SheetSizeDialog()
{
	DTP("TCADSheet::SheetSizeDialog()\n");

	CanvasSizeDialog(size,SetSheetSizeCB,this);
}

void
TCADSheet::Changed()
{
	int T=IsModified();
	
	DTP("TCADSheet::Changed()\n");

	if (BlockChange) return;
	ChangeCount++;
	if (IsModified() || T) SetTitle();

	SaveState();
}

void
TCADSheet::Saved()
{
	DTP("TCADSheet::Saved()\n");

	SavedCount = ChangeCount;
	SetTitle();
}

void
TCADSheet::SaveState()
{
	TCADMemorySaveStream *S;

	DTP("TCADSheet::SaveState()\n");

	if (UMAct < 0) UMAct = 0;	// First
	else {
		UMAct = (UMAct+1)%MAX_UNDO_LEVELS;
		if (UMAct == UMFirst) {
			UMFirst = (UMFirst+1)%MAX_UNDO_LEVELS;
		}
	}

	if (!UndoMem[UMAct]) UndoMem[UMAct] = new TCADMemory();

	S = new TCADMemorySaveStream(*UndoMem[UMAct]);
	SaveProperties(S);
	delete S;

	UMLast = UMAct;
}

void
TCADSheet::Undo()
{
	TCADLoadStream *S;

	DTP("TCADSheet::Undo()\n");

	if (UMAct == UMFirst) return;	// Nothing to undo

	BlockChange++;

	if (UMAct == 0) UMAct = MAX_UNDO_LEVELS-1;
	else UMAct--;

	// Delete current contents
	SelectAll();
	DeleteSelected();

	S = new TCADMemoryLoadStream(*UndoMem[UMAct]);
	S->KeepChanges = true;
	Load(S);
	PostLoad(S);
	delete S;

	ChangeCount--;
	SetTitle();

	DrawRect(NULL);

	BlockChange--;
}

void
TCADSheet::Redo()
{
	TCADLoadStream *S;

	DTP("TCADSheet::Redo()\n");

	if (UMAct == UMLast) return;	// Nothing to redo

	BlockChange++;

	UMAct = (UMAct+1)%MAX_UNDO_LEVELS;

	// Delete current contents
	SelectAll();
	DeleteSelected();

	S = new TCADMemoryLoadStream(*UndoMem[UMAct]);
	S->KeepChanges = true;
	Load(S);
	PostLoad(S);
	delete S;

	ChangeCount++;
	SetTitle();

	DrawRect(NULL);
	
	BlockChange--;
}

void
TCADSheet::EditCopy()
{
	TCADMemorySaveStream *S;

	DTP("TCADSheet::EditCopy()\n");

	S = new TCADMemorySaveStream(System->clipboard);
	S->SaveOnlySelected = true;
	SaveProperties(S);
	delete S;
}

void
TCADSheet::EditCut()
{
	DTP("TCADSheet::EditCut()\n");

	EditCopy();
	DeleteSelected();
}

void
TCADSheet::EditPaste()
{
	unsigned long OLUniq = LastUniq;

	DTP("TCADSheet::EditPaste()\n");

	TCADLoadStream *S;

	BlockChange++;

	UnselectAll();

	S = new TCADMemoryLoadStream(System->clipboard);
	S->RenumberUniq = true;
	S->KeepChanges = true;
	Load(S);
	PostLoad(S);
	delete S;

	BlockChange--;

	SelectHigher(OLUniq);

	Changed();

	DrawRect(NULL);
}

/*
 * Sends all selected objects to back
 */
void
TCADSheet::SendToBack()
{
	TCADObject *Obj;

	DTP("TCADSheet::SendToBack()\n");

	Obj = FirstSelected;
	while(Obj) {
		TCADGroup::SendToBack(Obj);
		Obj = Obj->NextSelected;
	}
}

/*
 * Brings all selected objects to front
 */
void
TCADSheet::BringToFront()
{
	TCADObject *Obj;

	DTP("TCADSheet::BringToFront()\n");

	Obj = FirstSelected;
	while(Obj) {
		TCADGroup::BringToFront(Obj);
		Obj = Obj->NextSelected;
	}
}


/*******************************************************************/

/*
 * TCADSheetSize implementation
 */

void
TCADSheetSize::Save(TCADSaveStream *Fl)
{
	DTP("TCADSheetSize::Save()\n");

  Fl->SaveObjectName("TCADSheetSize");
  SaveProperties(Fl);
  Fl->SaveEndObject();
}

void
TCADSheetSize::SaveProperties(TCADSaveStream *Fl)
{
	int T1,T2;

	DTP("TCADSheetSize::SaveProperties()\n");

  Fl->SaveInt("Width",width);
  Fl->SaveInt("Height",height);
  Fl->SaveInt("Metric",(int)metric);
  Fl->SaveInt("Constrain",constrainRatio);
}

void
TCADSheetSize::Load(TCADLoadStream *Fl)
{
  char *ID;

	DTP("TCADSheetSize::Load()\n");

  for(;;) {
    ID = Fl->ReadID();
    if (*ID == 0) break;

    if (!LoadProperty(Fl,ID)) Fl->Skip(); // Read rest
  }
}

char
TCADSheetSize::LoadProperty(TCADLoadStream *Fl,char *ID)
{
  char Result=1;

	DTP("TCADSheetSize::LoadProperty()\n");

  if (strcasecmp("Width",ID) == 0) width = Fl->LoadInt();
  else if (strcasecmp("Height",ID) == 0) height = Fl->LoadInt();
  else if (strcasecmp("Metric",ID) == 0) metric = (GtkMetricType)Fl->LoadInt();
  else if (strcasecmp("Constrain",ID) == 0) constrainRatio = Fl->LoadInt();
	else Result=0;

  return(Result);
}

